#ifndef _STRALLOC_H_
#define _STRALLOC_H_

#ifdef STRING_STATS
#define ADD_NEW_STRING(len, overhead) num_distinct_strings++; bytes_distinct_strings += len + 1; overhead_bytes += overhead
#define SUB_NEW_STRING(len, overhead) num_distinct_strings--; bytes_distinct_strings -= len + 1; overhead_bytes -= overhead

#define ADD_STRING(len) allocd_strings++; allocd_bytes += len + 1
#define ADD_STRING_SIZE(len) allocd_bytes += len
#define SUB_STRING(len) allocd_strings--; allocd_bytes -= len + 1

#else
/* Blazing fast macros :) */
#define ADD_NEW_STRING(x, y)
#define SUB_NEW_STRING(x, y)
#define ADD_STRING(x)
#define ADD_STRING_SIZE(x)
#define SUB_STRING(x)
#endif

typedef struct malloc_block_s {
    unsigned short size;
    unsigned short ref;
} malloc_block_t;

#define MSTR_REF(x) ((((malloc_block_t *)(x)) - 1)->ref)
#define MSTR_SIZE(x) ((svalue_strlen_size = MSTR_BLOCK(x)->size), ~svalue_strlen_size ? svalue_strlen_size : strlen((x)+0xffff)+0xffff)
#define MSTR_BLOCK(x) (((malloc_block_t *)(x)) - 1) 
#define MSTR_UPDATE_SIZE(x, y) do { ADD_STRING_SIZE(y - MSTR_SIZE(x)); MSTR_BLOCK(x)->size = (y > 0xffff ? 0xffff : y); } while (0)

#define FREE_MSTR(x) do { SUB_STRING(MSTR_SIZE(x)); SUB_NEW_STRING(MSTR_SIZE(x), sizeof(malloc_block_t)); FREE(MSTR_BLOCK(x)); } while (0)

/* This counts on some rather crucial alignment between malloc_block_t and
   block_t */
#define COUNTED_STRLEN(x) MSTR_BLOCK(x)->size
#define COUNTED_REF(x)    MSTR_BLOCK(x)->ref

typedef struct block_s {
    struct block_s *next;	/* next block in the hash chain */
#ifdef DEBUGMALLOC_EXTENSIONS
    int extra_ref;
#endif
    /* these two must be last */
    unsigned short size;	/* length of the string */
    unsigned short refs;	/* reference count    */
}       block_t;

#define NEXT(x) (x)->next
#define REFS(x) (x)->refs
#define EXTRA_REF(x) (x)->extra_ref
#define SIZE(x) (x)->size
#define BLOCK(x) (((block_t *)(x)) - 1)	/* pointer arithmetic */
#define STRING(x) ((char *)(x + 1))

#define SHARED_STRLEN(x) SIZE(BLOCK(x))

#define SVALUE_STRLEN(x) (((x)->subtype & STRING_COUNTED) ? \
			  COUNTED_STRLEN((x)->u.string) : \
			  strlen((x)->u.string))

/*
 * stralloc.c
 */
void init_strings PROT((void));
char *findstring PROT((char *));
char *make_shared_string PROT((char *));
char *ref_string PROT((char *));
void free_string PROT((char *));
void deallocate_string PROT((char *));
int add_string_status PROT((int));

char *extend_string PROT((char *, int));

extern unsigned short svalue_strlen_size;

#ifdef STRING_STATS
extern int num_distinct_strings;
extern int bytes_distinct_strings;
extern int allocd_strings;
extern int allocd_bytes;
extern int overhead_bytes;
#endif

#endif
